import json
from typing import Any

import aiohttp
from pydantic import Field

from fridonai_core.plugins.utilities.base import BaseUtility


class RemoteUtility(BaseUtility):
    request_url: str = Field(default="http://localhost:3000/data/executor", exclude=True)

    async def _get_remote_response(self, request):
        async with aiohttp.ClientSession() as session:
            async with session.post(self.request_url, json=request) as response:
                data = await response.json()
                if "statusCode" in data:
                    if 500 > data["statusCode"] >= 400:
                        return data.get("message", "Something went wrong!")
                    if data["statusCode"] >= 500:
                        return "Something went wrong! Please try again later."
                print("Got Response", data)
                return json.dumps(data["data"])

    async def arun(self, *args, **kwargs) -> dict | str | Any:
        request = await self._arun(*args, **kwargs)
        return await self._get_remote_response(request)



from langchain_core.messages import ToolMessage

from fridonai_core.graph.models import structured_final_response_model
from fridonai_core.graph.states import State


def prepare_plugin_agent(state):
    tool_call = state["messages"][-1].tool_calls[0]
    return {
        "messages": [
            state["messages"][-1],
            ToolMessage(tool_call_id=tool_call["id"], content=""),
        ]
    }


def handle_tool_error(state) -> dict:
    error = state.get("error")
    tool_calls = state["messages"][-1].tool_calls
    return {
        "messages": [
            ToolMessage(
                content=f"Error: {repr(error)}\n please fix your mistakes.",
                tool_call_id=tc["id"],
            )
            for tc in tool_calls
        ]
    }


def generate_structured_response(state: State):
    messages = state["messages"]
    for i in range(len(messages) - 1, -1, -1):
        if messages[i].type == "human":
            response = structured_final_response_model.invoke(messages[i:])
            return {"final_response": response}


def leave_tool(state) -> dict:
    messages = []
    if state["messages"][-1].tool_calls:
        messages.append(
            ToolMessage(
                content="Resuming dialog with the host assistant. Please reflect on the past conversation and assist the user as needed.",
                tool_call_id=state["messages"][-1].tool_calls[0]["id"],
            )
        )
    return {
        "messages": messages,
    }



from fridonai_core.plugins.base import BasePlugin
from fridonai_core.plugins.schemas import BaseToolInput
from fridonai_core.plugins.registry import ensure_plugin_registry

__all__ = ["BasePlugin", "BaseToolInput", "ensure_plugin_registry"]



from functools import reduce
from typing import Literal

from pydantic import BaseModel

from fridonai_core.plugins.tools import BaseTool


class BasePlugin(BaseModel):
    name: str
    description: str
    tools: list[BaseTool]
    owner: str = "2snYEzbMckwnv85MW3s2sCaEQ1wtKZv2cj9WhbmDuuRD"
    type: Literal["subscription", "nft"] = "subscription"
    price: float | int = 0
    image_url: str | None = None
    exclude: bool = False
    json_format: bool = False

    @property
    def examples(self) -> list[dict]:
        examples = reduce(lambda a, b: a + b, [tool.examples for tool in filter(lambda x: x.helper is False, self.tools)], [])
        return examples

    @property
    def request_examples(self) -> list[str]:
        return [example["request"] for example in self.examples]

    @property
    def examples_as_str(self) -> str:
        return "\n".join(self.request_examples)

    @property
    def slug(self) -> str:
        def slugify(text):
            import re
            import unicodedata
            text = unicodedata.normalize('NFKD', text)
            text = text.encode('ascii', 'ignore').decode('ascii')
            text = re.sub(r'[^a-zA-Z0-9\s-]', '', text)
            text = re.sub(r'[\s-]+', '-', text.strip().lower())
            return text
        return slugify(self.name)

    def full_description(self, tool_descriptions=True):
        description = self.name + " - " + self.description
        if tool_descriptions:
            description += "\n Here are following functionalities of this plugin: \n"
            for tool in filter(lambda x: x.helper is False, self.tools):
                description += "\t" + tool.name + ": " + tool.description + "\n"
        description += "Examples: \n" + self.examples_as_str
        return description

    @property
    def output_format(self):
        if self.json_format:
            return "\n After using tools and having everything together, and `plugin_status` going to be True, that means tool done successfully what asked to be done, copy whole tool's answer response as json string in CompleteTool's answer. If it didn't went good generate explanatory text from output. \n"
        else:
            return "\n After using tools and having everything together, combine all tool's answer responses in CompleteTool's answer as a text. \n"

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "slug": self.slug,
            "imageUrl": self.image_url,
            "description": self.description,
            "type": self.type,
            "owner": self.owner,
            "examples": self.examples,
            "functions": [tool.to_json() for tool in self.tools],
            "price": self.price
        }



from fridonai_core.graph.base import create_graph
from fridonai_core.graph.base import generate_response
from fridonai_core.graph.models import get_model

__all__ = ["create_graph", "generate_response", "get_model"]



import os
from typing import Any

from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_openai import ChatOpenAI
from pydantic import BaseModel, Field

from fridonai_core.plugins.utilities.base import BaseUtility


class LLMUtility(BaseUtility):
    llm_job_description: str = Field(
        description="Explicit description of the llm what it should do."
    )
    structured_output: type[BaseModel] | None = None

    async def arun(self, *args, **kwargs) -> BaseModel | dict | str | Any:
        placeholders = await self._arun(*args, **kwargs)

        prompt = PromptTemplate.from_template(self.llm_job_description)
        llm = ChatOpenAI(
            model="gpt-4o",
            temperature=0,
            openai_api_key=os.environ.get("OPENAI_API_KEY"),
            verbose=True,
        )

        if self.structured_output is not None:
            llm = llm.with_structured_output(self.structured_output)
            chain = prompt | llm | StrOutputParser()
        else:
            chain = prompt | llm

        result = await chain.ainvoke(placeholders)

        return result



from langchain_core.prompts import ChatPromptTemplate


def create_supervised_prompt() -> ChatPromptTemplate:
    return ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a Fridon ultimate companion for crypto users. You are very friendly, fun and make everything to have great conversation with the users."
                "Users may show up with bunch of different questions, you have to use appropriate tools to answer them, consider you are a supervisor for other assistants, tools, those"
                "are responsible for carrying out given tasks. \n"
                "So you are a supervisor tasked with managing a conversation between assistants and finally delivering the answer to the user. "
                "Your main role is to answer user's query, for that delegate tasks to the appropriate tools based on the user's query. For each question use just one tool."
                "The player is not aware of the different specialized assistants, so do not mention them; Don't mention that you are supervisor as well, for users you are Fridon, crypto companion."
                "Provide detailed and concise response, don't talk too much. Don't copy paste tools' responses except json responses and don't make up any information, avoid hallucinations. "
                "\nDon't make up any information just use provided information and messages. If some tool wasn't able to do its job say it don't pretend that everything is fine."
                "If you are confused and user's question is appropriate for several tools ask question to specify and make it clear."
                "\nIf the question is general and is about crypto, about coins, solana, bonk, blockchain protocols asking for some possible solutions use the `ToSolanaBonkEducatorPlugin` tool.\n"
                "Don't call the same assistant more than once."
                "When complete tool contains json data don't use it in text generation, imagine it doesn't exist at all. Just say that you've done your job is plugin_status is true otherwise if it's not done. Don't use any location for json data or offer any download or something like that!!!",
            ),
            ("placeholder", "{messages}"),
        ]
    )


def create_agent_prompt(
    name: str, description: str, output_format: str
) -> ChatPromptTemplate:
    return ChatPromptTemplate.from_messages(
        [
            (
                "system",
                f"You are a specialized assistant named '{name}'. For handling user's requests using the provided tools."
                "Use appropriate tools to satisfy all user's needs. "
                "\nWhen you think that you gathered all the necessary information, call the *CompleteTool* tool to let the primary assistant take control,"
                "But before that ensure that you called at least one tool. Don't make up information by yourself fist use some tools and then generate response."
                f"\nYour description: {description}",
            ),
            ("placeholder", "{messages}"),
            ("system", output_format),
        ]
    )




from typing import Optional
from pydantic import Field
from fridonai_core.plugins.utilities import BaseUtility
from fridonai_core.plugins.utilities.adapter_base import BaseAdapter

class BaseMockUtility(BaseUtility):
    async def arun(self, *args, **kwargs) -> str:
        pass
    model_config = {
        'ignored_types': (dict,)
    }

class BlockchainMockUtility(BaseMockUtility):
    communicator: Optional[BaseAdapter] = Field(default=None,exclude=True)

    async def arun(self, *args, **kwargs) -> str:
        return "Tx 12345 successfully completed!"

class RemoteMockUtility(BaseMockUtility):
    async def arun(self, *args, **kwargs) -> dict:
        return {"data": {"RSA": 23}}

class LLMMockUtility(BaseMockUtility):
    async def arun(self, *args, **kwargs) -> str:
        return "Fridon is Great"




import os 
from langchain_openai import ChatOpenAI

from fridonai_core.graph.tools import FinalResponse

chat_gpt_model = ChatOpenAI(
    model="gpt-4o", 
    temperature=0, 
    api_key=os.environ.get("OPENAI_API_KEY"), 
    verbose=True
)


def get_model(name="gpt"):
    return chat_gpt_model


def create_structured_output_model(output_format, model=chat_gpt_model):
    return model.with_structured_output(output_format)


structured_final_response_model = create_structured_output_model(FinalResponse)
